@@ -37,6 +37,7 @@
 #include <glib.h>
 #include <glib-object.h>
 #include "my-object.h"
+#include "test/lib/util.h"
 
 GMainLoop *loop = NULL;
 
@@ -56,6 +57,16 @@ static gboolean proxy_destroy_and_nameowner_complete = FALSE;
 static DBusGProxy *test_terminate_proxy1 = NULL;
 static DBusGProxy *test_terminate_proxy2 = NULL;
 
+static void
+cancel_exit_timeout (void)
+{
+  if (exit_timeout != 0)
+    {
+      g_source_remove (exit_timeout);
+      exit_timeout = 0;
+    }
+}
+
 #define lose(...) g_error (__VA_ARGS__)
 
 static void lose_gerror (const char *prefix, GError *error) G_GNUC_NORETURN;
@@ -72,6 +83,7 @@ timed_exit (gpointer loop)
 {
   g_print ("timed exit!\n");
   g_main_loop_quit (loop);
+  /* will repeat, so do not clear exit_timeout */
   return TRUE;
 }
 
@@ -81,7 +93,7 @@ proxy_destroyed_cb (DBusGProxy *proxy, gpointer user_data)
   proxy_destroyed = TRUE;
   if (proxy_destroy_and_nameowner && !proxy_destroy_and_nameowner_complete && await_terminating_service == NULL)
     {
-      g_source_remove (exit_timeout);
+      cancel_exit_timeout ();
       g_main_loop_quit (loop);
       proxy_destroy_and_nameowner_complete = TRUE;
     }
@@ -91,11 +103,14 @@ static void
 test_terminate_proxy1_destroyed_cb (DBusGProxy *proxy, gpointer user_data)
 {
    proxy_destroyed = TRUE;
+
   if (proxy_destroy_and_nameowner && !proxy_destroy_and_nameowner_complete && await_terminating_service == NULL)
     {
+      g_object_unref(test_terminate_proxy1);
+      test_terminate_proxy1 = NULL;
       g_object_unref(test_terminate_proxy2);
       test_terminate_proxy2 = NULL;
-      g_source_remove (exit_timeout);
+      cancel_exit_timeout ();
       g_main_loop_quit (loop);
       proxy_destroy_and_nameowner_complete = TRUE;
     }
@@ -119,13 +134,13 @@ name_owner_changed (DBusGProxy *proxy,
       await_terminating_service = NULL;
       if (proxy_destroy_and_nameowner && !proxy_destroy_and_nameowner_complete && proxy_destroyed)
 	{
-	  g_source_remove (exit_timeout);
+	  cancel_exit_timeout ();
 	  g_main_loop_quit (loop);
 	  proxy_destroy_and_nameowner_complete = TRUE;
 	} 
       else if (!proxy_destroy_and_nameowner)
 	{
-	  g_source_remove (exit_timeout);
+	  cancel_exit_timeout ();
 	  g_main_loop_quit (loop);
 	}
     }
@@ -141,7 +156,7 @@ foo_signal_handler (DBusGProxy  *proxy,
   g_print ("Got Foo signal\n");
 
   g_main_loop_quit (loop);
-  g_source_remove (exit_timeout);
+  cancel_exit_timeout ();
 }
 
 static void
@@ -155,7 +170,7 @@ frobnicate_signal_handler (DBusGProxy  *proxy,
   g_print ("Got Frobnicate signal\n");
 
   g_main_loop_quit (loop);
-  g_source_remove (exit_timeout);
+  cancel_exit_timeout ();
 }
 
 static void
@@ -180,7 +195,7 @@ frobnicate_signal_handler_compat (DBusGProxy  *proxy,
   g_print ("Got Frobnicate signal (compat)\n");
 
   g_main_loop_quit (loop);
-  g_source_remove (exit_timeout);
+  cancel_exit_timeout ();
 }
 
 static void
@@ -201,7 +216,7 @@ sig0_signal_handler (DBusGProxy  *proxy,
   g_print ("Got Sig0 signal\n");
 
   g_main_loop_quit (loop);
-  g_source_remove (exit_timeout);
+  cancel_exit_timeout ();
 }
 
 static void
@@ -221,7 +236,7 @@ sig1_signal_handler (DBusGProxy  *proxy,
   g_print ("Got Sig1 signal\n");
 
   g_main_loop_quit (loop);
-  g_source_remove (exit_timeout);
+  cancel_exit_timeout ();
 }
 
 static void
@@ -241,7 +256,7 @@ sig2_signal_handler (DBusGProxy  *proxy,
   g_print ("Got Sig2 signal\n");
 
   g_main_loop_quit (loop);
-  g_source_remove (exit_timeout);
+  cancel_exit_timeout ();
 }
 
 static DBusGProxyCall *echo_call;
@@ -270,7 +285,7 @@ echo_received_cb (DBusGProxy *proxy,
   g_print ("Async echo gave \"%s\"\n", echo_data); 
   g_free (echo_data);
   g_main_loop_quit (loop);
-  g_source_remove (exit_timeout);
+  cancel_exit_timeout ();
 }
 
 static void
@@ -294,7 +309,7 @@ increment_received_cb (DBusGProxy *proxy,
   
   g_print ("Async increment gave \"%d\"\n", val); 
   g_main_loop_quit (loop);
-  g_source_remove (exit_timeout);
+  cancel_exit_timeout ();
 }
 
 static void
@@ -310,7 +325,7 @@ increment_async_cb (DBusGProxy *proxy, guint val, GError *error, gpointer data)
 
   g_print ("(wrapped async) increment gave \"%d\"\n", val); 
   g_main_loop_quit (loop);
-  g_source_remove (exit_timeout);
+  cancel_exit_timeout ();
 }
 
 #define DBUS_TYPE_G_MAP_OF_VARIANT (dbus_g_type_get_map ("GHashTable", G_TYPE_STRING, G_TYPE_VALUE))
@@ -544,16 +559,11 @@ main (int argc, char **argv)
   loop = g_main_loop_new (NULL, FALSE);
 
   error = NULL;
-  connection = dbus_g_bus_get (DBUS_BUS_SESSION,
-                               &error);
+  connection = dbus_g_bus_get_private (DBUS_BUS_SESSION, NULL, &error);
+
   if (connection == NULL)
     lose_gerror ("Failed to open connection to bus", error);
 
-  /* should always get the same one */
-  g_assert (connection == dbus_g_bus_get (DBUS_BUS_SESSION, NULL));
-  g_assert (connection == dbus_g_bus_get (DBUS_BUS_SESSION, NULL));
-  g_assert (connection == dbus_g_bus_get (DBUS_BUS_SESSION, NULL));
-  
   /* Create a proxy object for the "bus driver" */
   
   driver = dbus_g_proxy_new_for_name (connection,
@@ -777,6 +787,7 @@ main (int argc, char **argv)
     lose ("Failed to begin Increment call");
 
   dbus_g_connection_flush (connection);
+  cancel_exit_timeout ();
   exit_timeout = g_timeout_add (5000, timed_exit, loop);
   g_main_loop_run (loop);
 
@@ -874,6 +885,7 @@ main (int argc, char **argv)
   if (!org_freedesktop_DBus_GLib_Tests_MyObject_increment_async (proxy, 42, increment_async_cb, NULL))
     lose_gerror ("Failed to complete (wrapped) Increment call", error);
   dbus_g_connection_flush (connection);
+  cancel_exit_timeout ();
   exit_timeout = g_timeout_add (5000, timed_exit, loop);
   g_main_loop_run (loop);
 
@@ -1031,6 +1043,8 @@ main (int argc, char **argv)
       lose_gerror ("Failed to complete (wrapped) zero-length recursive1 call", error);
     if (arraylen != 0)
       lose ("(wrapped) zero-length recursive1 call returned invalid length %u", arraylen);
+
+    g_array_unref (array);
   }
 
   {
@@ -1057,6 +1071,8 @@ main (int argc, char **argv)
       lose_gerror ("Failed to complete (wrapped) recursive1 call", error);
     if (arraylen != 5)
       lose ("(wrapped) recursive1 call returned invalid length %u", arraylen);
+
+    g_array_unref (array);
   }
 
   {
@@ -1291,6 +1307,7 @@ main (int argc, char **argv)
       lose_gerror ("Failed to complete RecArrays call", error);
     g_free (g_ptr_array_index (in_array, 0));
     g_free (g_ptr_array_index (in_array, 1));
+    g_ptr_array_free (in_array, TRUE);
 
     g_assert (out_array);
     g_assert (out_array->len == 2);
@@ -1392,6 +1409,8 @@ main (int argc, char **argv)
       lose_gerror ("Failed to complete (wrapped) GetValue call", error);
     if (val != 3)
       lose ("(wrapped) GetValue returned invalid value %d", val);
+
+    g_object_unref (ret_proxy);
   }
 
   run_mainloop ();
@@ -1535,8 +1554,6 @@ main (int argc, char **argv)
 
     g_hash_table_destroy (table);
     g_hash_table_destroy (ret_table);
-
-    g_mem_profile ();
   }
 
   for (i=0; i<3; i++)
@@ -1583,8 +1600,6 @@ main (int argc, char **argv)
 
     g_hash_table_destroy (table);
     g_hash_table_destroy (ret_table);
-
-    g_mem_profile ();
   }
 
   for (i=0; i<3; i++)
@@ -1638,8 +1653,6 @@ main (int argc, char **argv)
 
     g_boxed_free (dbus_g_type_get_map ("GHashTable", DBUS_TYPE_G_OBJECT_PATH,
           dbus_g_type_get_collection ("GPtrArray", DBUS_TYPE_G_OBJECT_PATH)), ret_table);
-
-    g_mem_profile ();
   }
 
 
@@ -1660,6 +1673,7 @@ main (int argc, char **argv)
 
   
   dbus_g_connection_flush (connection);
+  cancel_exit_timeout ();
   exit_timeout = g_timeout_add (5000, timed_exit, loop);
   g_main_loop_run (loop);
 
@@ -1672,6 +1686,7 @@ main (int argc, char **argv)
     lose_gerror ("Failed to complete EmitFrobnicate call", error);
   
   dbus_g_connection_flush (connection);
+  cancel_exit_timeout ();
   exit_timeout = g_timeout_add (5000, timed_exit, loop);
   g_main_loop_run (loop);
 
@@ -1692,8 +1707,6 @@ main (int argc, char **argv)
   if (proxy == NULL)
     lose_gerror ("Failed to create proxy for name owner", error);
 
-  my_object_register_marshallers ();
-
   dbus_g_proxy_add_signal (proxy, "Sig0", G_TYPE_STRING, G_TYPE_INT, G_TYPE_STRING, G_TYPE_INVALID);
   dbus_g_proxy_add_signal (proxy, "Sig1", G_TYPE_STRING, G_TYPE_VALUE, G_TYPE_INVALID);
   dbus_g_proxy_add_signal (proxy, "Sig2", DBUS_TYPE_G_STRING_STRING_HASHTABLE, G_TYPE_INVALID);
@@ -1712,8 +1725,10 @@ main (int argc, char **argv)
   dbus_g_proxy_call_no_reply (proxy, "EmitSignals", G_TYPE_INVALID);
 
   dbus_g_connection_flush (connection);
+  cancel_exit_timeout ();
   exit_timeout = g_timeout_add (5000, timed_exit, loop);
   g_main_loop_run (loop);
+  cancel_exit_timeout ();
   exit_timeout = g_timeout_add (5000, timed_exit, loop);
   g_main_loop_run (loop);
 
@@ -1726,6 +1741,7 @@ main (int argc, char **argv)
   dbus_g_proxy_call_no_reply (proxy, "EmitSignal2", G_TYPE_INVALID);
   dbus_g_connection_flush (connection);
 
+  cancel_exit_timeout ();
   exit_timeout = g_timeout_add (5000, timed_exit, loop);
   g_main_loop_run (loop);
 
@@ -1737,12 +1753,16 @@ main (int argc, char **argv)
   dbus_g_proxy_call_no_reply (proxy, "EmitSignals", G_TYPE_INVALID);
 
   dbus_g_connection_flush (connection);
+  cancel_exit_timeout ();
   exit_timeout = g_timeout_add (5000, timed_exit, loop);
   g_main_loop_run (loop);
+  cancel_exit_timeout ();
   exit_timeout = g_timeout_add (5000, timed_exit, loop);
   g_main_loop_run (loop);
+  cancel_exit_timeout ();
   exit_timeout = g_timeout_add (5000, timed_exit, loop);
   g_main_loop_run (loop);
+  cancel_exit_timeout ();
   exit_timeout = g_timeout_add (5000, timed_exit, loop);
   g_main_loop_run (loop);
 
@@ -1766,6 +1786,7 @@ main (int argc, char **argv)
 		    NULL);
 
   dbus_g_connection_flush (connection);
+  cancel_exit_timeout ();
   exit_timeout = g_timeout_add (5000, timed_exit, loop);
   g_main_loop_run (loop);
 
@@ -1775,7 +1796,8 @@ main (int argc, char **argv)
     lose ("Didn't get proxy_destroyed");
   g_print ("Proxy destroyed successfully\n");
 
-  /* Don't need to unref, proxy was destroyed */
+  /* "destroy" does not mean last-unref! */
+  g_object_unref (proxy);
 
   run_mainloop ();
 
@@ -1840,6 +1862,7 @@ main (int argc, char **argv)
   n_times_frobnicate_received = 0;
 
   dbus_g_connection_flush (connection);
+  cancel_exit_timeout ();
   exit_timeout = g_timeout_add (5000, timed_exit, loop);
   g_main_loop_run (loop);
 
@@ -1851,6 +1874,7 @@ main (int argc, char **argv)
   await_terminating_service = "org.freedesktop.DBus.GLib.TestService";
   dbus_g_proxy_call_no_reply (proxy, "Terminate", G_TYPE_INVALID);
   dbus_g_connection_flush (connection);
+  cancel_exit_timeout ();
   exit_timeout = g_timeout_add (5000, timed_exit, loop);
   g_main_loop_run (loop);
   if (await_terminating_service != NULL)
@@ -1870,6 +1894,7 @@ main (int argc, char **argv)
     g_usleep (8 * G_USEC_PER_SEC);
 
   dbus_g_connection_flush (connection);
+  cancel_exit_timeout ();
   exit_timeout = g_timeout_add (5000, timed_exit, loop);
   g_main_loop_run (loop);
 
@@ -1898,6 +1923,7 @@ main (int argc, char **argv)
     lose_gerror ("Failed to complete EmitFrobnicate call", error);
 
   dbus_g_connection_flush (connection);
+  cancel_exit_timeout ();
   exit_timeout = g_timeout_add (5000, timed_exit, loop);
   g_main_loop_run (loop);
 
@@ -2026,7 +2052,8 @@ main (int argc, char **argv)
 
     if (!(found_introspectable && found_myobject && found_properties))
       lose ("Missing interface"); 
-    g_free (node);
+
+    node_info_unref (node);
   }
   g_free (v_STRING_2);
 
@@ -2336,6 +2363,7 @@ main (int argc, char **argv)
 		    NULL);
 
   dbus_g_connection_flush (connection);
+  cancel_exit_timeout ();
   exit_timeout = g_timeout_add (5000, timed_exit, loop);
   g_main_loop_run (loop);
 
@@ -2367,11 +2395,20 @@ main (int argc, char **argv)
       lose_gerror ("Failed to complete (wrapped) DoNothing call", error);
     
     g_object_unref (G_OBJECT (proxy));
+
+    test_run_until_disconnected (privconn, NULL);
+    dbus_g_connection_unref (privconn);
   }
 
   g_object_unref (G_OBJECT (driver));
 
+  test_run_until_disconnected (connection, NULL);
+  dbus_g_connection_unref (connection);
+
   g_print ("Successfully completed %s\n", argv[0]);
 
+  dbus_shutdown ();
+  g_main_loop_unref (loop);
+
   return 0;
 }
